/* we have to include loader2.h before windows.h */
#include "loader2.h"
#include <windows.h>
#include <stdio.h>

#ifndef JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE
	/* from http://msdn.microsoft.com/en-us/library/windows/desktop/ms684147(v=vs.85).aspx */
	#define JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE 0x00002000
#endif

static int CreateProcessInJob(char *cmd, HANDLE *stdhandle, PROCESS_INFORMATION *pi)
{
	STARTUPINFO si;
	JOBOBJECT_EXTENDED_LIMIT_INFORMATION li;
	HANDLE job;
	int hasjobs = 1;

	job = CreateJobObject(NULL, NULL);
	if (job == NULL) {
		fprintf(stderr, "CreateJobObject error: %d\n", GetLastError());
		fprintf(stderr, "No jobs used - in case of a crash, the subprocess may not quit\n");
		hasjobs = 0;
	}

	if (hasjobs) {
		ZeroMemory(&li, sizeof(li));
		li.BasicLimitInformation.LimitFlags = JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE;
		if (SetInformationJobObject(job, JobObjectExtendedLimitInformation, &li, sizeof(li)) == 0) {
			fprintf(stderr, "SetInformationJobObject error: %d\n", GetLastError());
			fprintf(stderr, "No jobs used - in case of a crash, the subprocess may not quit\n");
			hasjobs = 0;
		}
	}

	ZeroMemory(pi, sizeof(PROCESS_INFORMATION));
	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);
	si.dwFlags = STARTF_USESTDHANDLES;
	si.hStdInput = stdhandle[0];
	si.hStdOutput = stdhandle[1];
	si.hStdError = stdhandle[2];

	if (CreateProcess(NULL, cmd, NULL, NULL, TRUE, CREATE_SUSPENDED, NULL, NULL, &si, pi) == 0) {
		fprintf(stderr, "CreateProcess error: %d\n", GetLastError());
		return -1;
	}

	if (hasjobs) {
		if (AssignProcessToJobObject(job, pi->hProcess) == 0) {
			fprintf(stderr, "AssignProcessToJobObject error: %d\n", GetLastError());
			fprintf(stderr, "No jobs used - in case of a crash, the subprocess may not quit\n");
		}
	}

	return 0;
}

static int RunProcess(PROCESS_INFORMATION *pi)
{
	if (ResumeThread(pi->hThread) == -1) {
		fprintf(stderr, "ResumeThread error: %d\n", GetLastError());
		return -1;
	}

	return 0;
}

static int CallRemoteFunction(PROCESS_INFORMATION *pi, char *dllname, char *funcname, void *param, int len, int errvalue)
{
	HMODULE dll;
	HANDLE thread;
	void *func, *mem = NULL;
	int ret;

	dll = LoadLibrary(dllname);
	if (dll == NULL) {
		fprintf(stderr, "LoadLibrary error (%s): %d\n", dllname, GetLastError());
		return errvalue;
	}

	func = GetProcAddress(dll, funcname);
	if (func == NULL) {
		fprintf(stderr, "GetProcAddress error (%s): %d\n", funcname, GetLastError());
		return errvalue;
	}

	if (param != NULL) {
		mem = VirtualAllocEx(pi->hProcess, NULL, len, MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE);
		if (mem == NULL) {
			fprintf(stderr, "VirtualAllocEx error: %d\n", GetLastError());
			return errvalue;
		}

		if (WriteProcessMemory(pi->hProcess, mem, param, len, NULL) == 0) {
			fprintf(stderr, "WriteProcessMemory error: %d\n", GetLastError());
			return errvalue;
		}
	}

	thread = CreateRemoteThread(pi->hProcess, NULL, 0, (LPTHREAD_START_ROUTINE)func, mem, 0, NULL);
	if (thread == NULL) {
		fprintf(stderr, "CreateRemoteThread error: %d\n", GetLastError());
		return errvalue;
	}

	if (WaitForSingleObject(thread, INFINITE) != WAIT_OBJECT_0) {
		fprintf(stderr, "WaitForSingleObject error: %d\n", GetLastError());
		return errvalue;
	}

	if (GetExitCodeThread(thread, (void *)&ret) == 0) {
		fprintf(stderr, "GetExitCodeThread error: %d\n", GetLastError());
		return errvalue;
	}

	CloseHandle(thread);
	if (param != NULL) {
		VirtualFreeEx(pi->hProcess, mem, 0, MEM_RELEASE);
	}
	FreeLibrary(dll);

	return ret;
}

int RunProcessWithFds(char *cmd, loader2_param_t *param, int len, PROCESS_INFORMATION *pi)
{
	int ret, i;
	HANDLE stdhandle[3] = {NULL, NULL, NULL};

	for (i = 0; i < param->fdn; i++) {
		if (SetHandleInformation(param->fds[i].handle, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT) == 0) {
			fprintf(stderr, "SetHandleInformation error: %d\n", GetLastError());
			return -1;
		}
		if (param->fds[i].fd < 3) {
			stdhandle[param->fds[i].fd] = param->fds[i].handle;
		}
	}

	ret = CreateProcessInJob(cmd, stdhandle, pi);
	if (ret != 0) {
		fprintf(stderr, "CreateProcessInJob error: %d\n", ret);
		return -1;
	}

	if (CallRemoteFunction(pi, "kernel32.dll", "LoadLibraryA", "libporty_abgproc_helper.dll", strlen("libporty_abgproc_helper.dll")+1, 0) == 0) {
		ret = CallRemoteFunction(pi, "kernel32.dll", "GetLastError", NULL, 0, -12345);
		fprintf(stderr, "CallRemoteFunction error (LoadLibraryA): %d\n", ret);
		return -1;
	}

	ret = CallRemoteFunction(pi, "libporty_abgproc_helper.dll", "Initialize", param, len, -1);
	if (ret != 0) {
		fprintf(stderr, "CallRemoteFunction error (my func): %d\n", ret);
		return -1;
	}

	return RunProcess(pi);
}
